"""Transformer for preprocessing input data and postprocessing output data"""

from synthesis.transformers.generalization import *
from synthesis.transformers.sequence import GeneralizeDateSequence
from synthesis.transformers.deidentification import *

__all__ = [
    'GeneralizeContinuous',
    'GeneralizeCategorical',
    'GeneralizeSchematic',
    'GeneralizeDateSequence'
]



"""Functions to help perform an experimental run of synthesis"""
from itertools import product
from pathlib import Path

from synthesis.synthesizers.privbayes import PrivBayes
from synthesis.synthesizers.marginal import MarginalSynthesizer

def synthesis_experiment(X, X_name, synthesizers=None, epsilon=None, n_records_synth=None, path=None, verbose=2):
    if synthesizers is None:
        synthesizers = [MarginalSynthesizer, PrivBayes]

    if not isinstance(path, Path):
        path = Path(path)
    model_path = path / 'output_model'
    data_path = path / 'output_data'

    if epsilon is None:
        epsilon = [0.01, 0.1, 1]

    for sts, e in product(synthesizers, epsilon):
        synthesizer = sts(epsilon=e, verbose=verbose)
        if isinstance(synthesizer, MarginalSynthesizer):
            synthesizer.verbose = 0

        synthesizer.fit(X)
        # Xs = synthesizer.sample()
        # synthesizer.save(model_path)
        print('Synthesis complete for: {}'.format(sts))

    return synthesizer


"""
Synthetic Data Generation via marginal distributions.
"""

import numpy as np
import pandas as pd
from synthesis.synthesizers._base import BaseDPSynthesizer
from synthesis.synthesizers.utils import dp_marginal_distribution, uniform_distribution


class MarginalSynthesizer(BaseDPSynthesizer):
    """Synthetic Data Generation via marginal distributions.

    Aims to only preserve the marginal distributions of each column independently. Does
    not preserve the statistical patterns between columns. Will work with any dataset
    that fits into memory."""

    def __init__(self, epsilon=1.0, verbose=True):
        super().__init__(epsilon=epsilon, verbose=verbose)

    def fit(self, data):
        data = self._check_input_data(data)
        self._check_init_args()
        # divide epsilon budget over each column
        local_epsilon = self.epsilon / data.shape[1]

        # note that in Python 3 dicts remember insertion order - thus no need to use ordered dict
        self.model_ = {}
        for c in data.columns:
            marginal = dp_marginal_distribution(data[c], local_epsilon)
            self.model_[c] = dict(zip(marginal.as_dict()['states'][c], marginal.as_dict()['data']))
            if self.verbose:
                print('Marginal fitted: {}'.format(c))
        return self

    def sample(self, n_records=None):
        self._check_is_fitted()
        n_records = n_records or self.n_records_fit_

        data_synth = {}
        # sample columns independently from marginal distributions
        for c in self.columns_:
            column_values = list(self.model_[c].keys())
            column_value_probabilities = list(self.model_[c].values())
            column_sampled = np.random.choice(column_values, p=column_value_probabilities, size=n_records, replace=True)
            data_synth[c] = column_sampled
            if self.verbose:
                print('Column sampled: {}'.format(c))
        data_synth = pd.DataFrame(data_synth)
        data_synth = self._check_output_data(data_synth)
        return data_synth


class UniformSynthesizer(MarginalSynthesizer):
    """Synthetic Data Generation via uniform distribution.
    """

    def __init__(self, epsilon=1.0, verbose=True):
        super().__init__(epsilon=epsilon, verbose=verbose)

    def fit(self, data):
        data = self._check_input_data(data)
        self._check_init_args()

        self.model_ = {}
        for c in data.columns:
            uniform = uniform_distribution(data[c])
            self.model_[c] = dict(zip(uniform.as_dict()['states'][c], uniform.as_dict()['data']))
            if self.verbose:
                print('Uniform fitted: {}'.format(c))
        return self




"""
Utility evaluator. Comparing a reference dataset to 1 or more target datasets.
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from pathlib import Path

import synthesis.evaluation.metrics as metrics
from synthesis.evaluation._base import BaseMetric, COLOR_PALETTE



DEFAULT_METRICS = {
    'average_js_distance': metrics.MarginalComparison(),
    'pairwise_correlation_distance': metrics.AssociationsComparison()
}

class SyntheticDataEvaluator(BaseMetric):
    """Class to compare synthetic data to the original"""
    def __init__(self, metrics=None):
        """Choose which metrics to compute"""
        self.metrics = metrics

    def fit(self, data_original, data_synthetic):
        self._check_input_args()
        data_original, data_synthetic = self._check_input_data(data_original, data_synthetic)

        for name, metric in self.metrics.items():
            metric.fit(data_original, data_synthetic)
        return self

    def score(self):
        scores = {}
        for name, metric in self.metrics.items():
            scores[name] = metric.score()
        return scores

    def plot(self):
        for name, metric in self.metrics.items():
            metric.plot()

    def _check_input_args(self):
        if self.metrics is not None:
            for name, metric in self.metrics.items():
                if not isinstance(metric, BaseMetric):
                    raise ValueError("Input metric {} should subclass synthesis.evaluation._base.BaseMetric".format(metric))
        else:
            self.metrics = DEFAULT_METRICS


class OriginalDataEvaluator():
    """Class to evaluate input dataframe"""
    def __init__(self, cardinality_threshold=50, rare_category_threshold=0.05):
        self.cardinality_threshold = cardinality_threshold
        self.rare_category_threshold = rare_category_threshold

    def fit(self, data):
        self.stats_ = {}
        self.stats_['columns_high_cardinality'] = self.get_high_cardinality_columns(data, self.cardinality_threshold)
        self.stats_['rare_column_categories'] = self.get_rare_column_categories(data, self.rare_category_threshold)
        return self

    def plot(self, data, normalize=True):
        column_names = data.columns
        fig, ax = plt.subplots(len(column_names), 1, figsize=(8, len(column_names) * 4))

        for idx, col in enumerate(column_names):
            column_value_counts = data.value_counts(normalize=normalize)

            bar_position = np.arange(len(column_value_counts.values))
            bar_width = 0.5

            ax[idx].bar(x=bar_position, height=column_value_counts.values,
                            color=COLOR_PALETTE[0], label='original', width=bar_width)

            ax[idx].set_xticks(bar_position + bar_width / 2)
            if len(column_value_counts.values) <= 20:
                ax[idx].set_xticklabels(column_value_counts.keys(), rotation=25)
            else:
                ax[idx].set_xticklabels('')

            title = r"$\bf{" + col + "}$"
            ax[idx].set_title(title)
            if normalize:
                ax[idx].set_ylabel('Probability')
            else:
                ax[idx].set_ylabel('Count')

            ax[idx].legend()
        fig.tight_layout()
        plt.show()

    @staticmethod
    def get_high_cardinality_columns(data, threshold):
        """Get features with more unique values than the specified threshold."""
        return data.columns[data.nunique() > threshold].tolist()

    @staticmethod
    def get_rare_column_categories(data, threshold):
        """Get rare categories per column"""
        rare_categories = {}
        for c in data.columns:
            rare_categories[c] = [k for k, v in data[c].value_counts(normalize=True).items() if v < threshold]
        return rare_categories




"""
Synthetic data generation via Bayesian Networks

Based on following paper

Zhang J, Cormode G, Procopiuc CM, Srivastava D, Xiao X.
PrivBayes: Private Data Release via Bayesian Networks. (2017)
"""

import numpy as np
import pandas as pd
import random
import dill
from sklearn.metrics import mutual_info_score
from collections import namedtuple
from joblib import Parallel, delayed


import synthesis.synthesizers.utils as utils
from synthesis.synthesizers._base import BaseDPSynthesizer
from thomas.core import BayesianNetwork
from diffprivlib.mechanisms import Exponential

APPair = namedtuple('APPair', ['attribute', 'parents'])


class PrivBayes(BaseDPSynthesizer):
    """PrivBayes: Private Data Release via Bayesian Networks (Zhang et al 2017)

    Version:
    - vanilla encoding
    - mutual information as score function

    Extended:
    - Ability to initialize network

    Default hyperparameters set according to paper recommendations
    """

    def __init__(self, epsilon=1.0, theta_usefulness=4, epsilon_split=0.3,
                 score_function='R', network_init=None, n_cpus=None, verbose=True):
        super().__init__(epsilon=epsilon, verbose=verbose)
        self.theta_usefulness = theta_usefulness
        self.epsilon_split = epsilon_split  # also called Beta in paper
        self.score_function = score_function  # choose between 'R' and 'MI'
        self.network_init = network_init
        self.n_cpus = n_cpus

    def fit(self, data):
        data = self._check_input_data(data)
        self._check_init_args()

        self._greedy_bayes(data)
        self._compute_conditional_distributions(data)
        self.model_ = BayesianNetwork.from_CPTs('PrivBayes', self.cpt_.values())
        return self

    def _check_init_args(self):
        super()._check_init_args()
        self._check_score_function()

    def _check_score_function(self):
        """Checks input score function and sets sensitivity"""
        if self.score_function.upper() not in ['R', 'MI']:
            raise ValueError("Score function must be 'R' or 'MI'")

        if self.score_function.upper() == 'R':
            self._score_sensitivity = (3 / self.n_records_fit_) + (2 / self.n_records_fit_**2)

        # note: for simplicity we assume that all APPairs are non-binary, which is the upperbound of MI sensitivity
        elif self.score_function.upper() == 'MI':
            self._score_sensitivity = (2 / self.n_records_fit_) * np.log((self.n_records_fit_ + 1) / 2) + \
                              (((self.n_records_fit_ - 1) / self.n_records_fit_) *
                               np.log((self.n_records_fit_ + 1) / (self.n_records_fit_ - 1)))

    def sample(self, n_records=None):
        self._check_is_fitted()
        n_records = n_records or self.n_records_fit_

        data_synth = self._generate_data(n_records)
        data_synth = self._check_output_data(data_synth)

        if self.verbose:
            print("\nSynthetic Data Generated\n")
        return data_synth

    def _greedy_bayes(self, data):
        nodes, nodes_selected = self._init_network(data)

        # normally len(nodes) - 1, unless user initialized part of the network
        self._n_nodes_dp_computed = len(nodes) - len(nodes_selected)

        for i in range(len(nodes_selected), len(nodes)):
            if self.verbose:
                print("{}/{} - Evaluating next attribute to add to network".format(i + 1, len(self.columns_)))

            nodes_remaining = nodes - nodes_selected

            # select ap_pair candidates
            ap_pairs = []
            for node in nodes_remaining:
                max_domain_size = self._max_domain_size(data, node)
                max_parent_sets = self._max_parent_sets(data, nodes_selected, max_domain_size)

                # empty set - domain size of node violates theta_usefulness
                if len(max_parent_sets) == 0:
                    ap_pairs.append(APPair(node, parents=None))
                # [empty set] - no parents found that meets domain size restrictions
                elif len(max_parent_sets) == 1 and len(max_parent_sets[0]) == 0:
                    ap_pairs.append(APPair(node, parents=None))
                else:
                    ap_pairs.extend([
                        APPair(node, parents=tuple(p)) for p in max_parent_sets
                    ])
            if self.verbose:
                print("Number of AttributeParentPair candidates: {}".format(len(ap_pairs)))

            scores = self._compute_scores(data, ap_pairs)
            sampled_pair = self._exponential_mechanism(ap_pairs, scores)

            if self.verbose:
                print("Selected attribute: '{}' - with parents: {}\n".format(sampled_pair.attribute, sampled_pair.parents))
            nodes_selected.add(sampled_pair.attribute)
            self.network_.append(sampled_pair)
        if self.verbose:
            print("Learned Network Structure\n")
        return self

    def _max_domain_size(self, data, node):
        """Computes the maximum domain size a node can have to satisfy theta-usefulness"""
        node_cardinality = utils.cardinality(data[node])
        max_domain_size = (self.n_records_fit_ * (1 - self.epsilon_split) * self.epsilon) / \
                          (2 * len(self.columns_) * self.theta_usefulness * node_cardinality)
        return max_domain_size

    def _max_parent_sets(self, data, v, max_domain_size):
        """Refer to algorithm 5 in paper - max parent set is 1) theta-useful and 2) maximal."""
        if max_domain_size < 1:
            return set()
        if len(v) == 0:
            return [set()]

        x = np.random.choice(tuple(v))
        x_domain_size = utils.cardinality(data[x])
        x = {x}

        v_without_x = v - x

        parent_sets1 = self._max_parent_sets(data, v_without_x, max_domain_size)
        parent_sets2 = self._max_parent_sets(data, v_without_x, max_domain_size / x_domain_size)

        for z in parent_sets2:
            if z in parent_sets1:
                parent_sets1.remove(z)
            parent_sets1.append(z.union(x))
        return parent_sets1

    def _init_network(self, X):
        self._binary_columns = [c for c in X.columns if X[c].unique().size <= 2]
        nodes = set(X.columns)

        if self.network_init is not None:
            nodes_selected = set(n.attribute for n in self.network_init)
            # print("Pre-defined network init: {}".format(self.network_))
            for i, pair in enumerate(self.network_init):
                if self.verbose:
                    print("{}/{} - init node {} - with parents: {}".format(i + 1, len(nodes),
                                                                           pair.attribute, pair.parents))
            self.network_ = self.network_init.copy()
            return nodes, nodes_selected

        # if set_network is not called we start with a random first node
        self.network_ = []
        nodes_selected = set()

        root = np.random.choice(tuple(nodes))
        self.network_.append(APPair(attribute=root, parents=None))
        nodes_selected.add(root)
        if self.verbose:
            print("1/{} - Root of network: {}\n".format(X.shape[1], root))
        return nodes, nodes_selected

    def set_network(self, network):
        assert [isinstance(n, APPair) for n in network], "input network does not consists of APPairs"
        self.network_init = network
        return self

    def _compute_scores(self, data, ap_pairs):
        """Compute score for all ap_pairs"""
        if self.n_cpus:
            scores = Parallel(n_jobs=self.n_cpus)(delayed(self.r_score)(data, pair.attribute, pair.parents) for pair in ap_pairs)
        else:
            scores = [self.r_score(data, pair.attribute, pair.parents) for pair in ap_pairs]
        return scores

    # def _compute_scores(self, data, ap_pairs):
    #     """Compute score for all ap_pairs"""
    #     scores = np.empty(len(ap_pairs))
    #
    #     for idx, pair in enumerate(ap_pairs):
    #         if pair.parents is None:
    #             scores[idx] = 0
    #         elif self.score_function == 'R':
    #                 scores[idx] = self.r_score(data, pair.attribute, pair.parents)
    #         elif self.score_function == 'MI':
    #             scores[idx] = self.mi_score(data, pair.attribute, pair.parents)
    #     return scores

    def _exponential_mechanism(self, ap_pairs, scores):
        """select APPair with exponential mechanism"""
        local_epsilon = self.epsilon * self.epsilon_split / self._n_nodes_dp_computed
        dp_mech = Exponential(epsilon=local_epsilon, sensitivity=self._score_sensitivity,
                              utility=list(scores), candidates=ap_pairs)
        sampled_pair = dp_mech.randomise()
        return sampled_pair

    def _compute_conditional_distributions(self, data):
        self.cpt_ = dict()

        local_epsilon = self.epsilon * (1 - self.epsilon_split) / len(self.columns_)

        for idx, pair in enumerate(self.network_):
            if pair.parents is None:
                attributes = [pair.attribute]
            else:
                attributes = [*pair.parents, pair.attribute]

            cpt_size = utils.cardinality(data[attributes])
            if self.verbose:
                print('Learning conditional probabilities: {} - with parents {} '
                      '~ estimated size: {}'.format(pair.attribute, pair.parents, cpt_size))

            dp_cpt = utils.dp_conditional_distribution(data[attributes], epsilon=local_epsilon)
            self.cpt_[pair.attribute] = dp_cpt
        return self

    def _generate_data(self, n_records):
        data_synth = np.empty([n_records, len(self.columns_)], dtype=object)

        for i in range(n_records):
            if self.verbose:
                print("\r", end='')
                print('Number of records generated: {} / {}'.format(i + 1, n_records), end='', flush=True)
            record = self._sample_record()
            data_synth[i] = list(record.values())

        # numpy.array to pandas.DataFrame with original column ordering
        data_synth = pd.DataFrame(data_synth, columns=[c.attribute for c in self.network_])
        return data_synth

    def _sample_record(self):
        """samples a value column for column by conditioning for parents"""
        record = {}
        for col_idx, pair in enumerate(self.network_):
            node = self.model_[pair.attribute]
            node_cpt = node.cpt
            node_states = node.states

            if node.conditioning:
                parent_values = [record[p] for p in node.conditioning]
                node_probs = node_cpt[tuple(parent_values)]

            else:
                node_probs = node_cpt.values
            # use random.choices over np.random.choice as np coerces, e.g. sample['nan', 1, 3.0] -> '1' (int to string)
            sampled_node_value = random.choices(node_states, weights=node_probs, k=1)[0] # returns list

            record[node.name] = sampled_node_value
        return record

    @staticmethod
    def mi_score(data, columns_a, columns_b):
        columns_a = utils._ensure_arg_is_list(columns_a)
        columns_b = utils._ensure_arg_is_list(columns_b)

        data_a = data[columns_a].squeeze()
        if len(columns_b) == 1:
            data_b = data[columns_b].squeeze()
        else:
            data_b = data.loc[:, columns_b].apply(lambda x: ' '.join(x.values), axis=1).squeeze()
        return mutual_info_score(data_a, data_b)

    @staticmethod
    def mi_score_thomas(data, columns_a, columns_b):
        columns_a = utils._ensure_arg_is_list(columns_a)
        columns_b = utils._ensure_arg_is_list(columns_b)

        prob_a = utils.compute_distribution(data[columns_a])
        prob_b = utils.compute_distribution(data[columns_b])
        prob_joint = utils.compute_distribution(data[columns_a + columns_b])

        # todo: pull-request thomas to add option to normalize to remove 0's
        # align
        prob_div = prob_joint / (prob_b * prob_a)
        prob_joint, prob_div = prob_joint.extend_and_reorder(prob_joint, prob_div)

        # remove zeros as this will result in issues with log
        prob_joint = prob_joint.values[prob_joint.values != 0]
        prob_div = prob_div.values[prob_div.values != 0]
        mi = np.sum(prob_joint * np.log(prob_div))
        # mi = np.sum(p_nodeparents.values * np.log(p_nodeparents / (p_parents * p_node)))
        return mi


    @staticmethod
    def r_score(data, columns_a, columns_b):
        """An alternative score function to mutual information with lower sensitivity - can be used on non-binary domains.
        Relies on the L1 distance from a joint distribution to a joint distributions that minimizes mutual information.
        Refer to Lemma 5.2
        """
        if columns_b is None:
            return 0
        columns_a = utils._ensure_arg_is_list(columns_a)
        columns_b = utils._ensure_arg_is_list(columns_b)

        # compute distribution that minimizes mutual information
        prob_a = utils.compute_distribution(data[columns_a])
        prob_b = utils.compute_distribution(data[columns_b])
        prob_independent = prob_b * prob_a

        # compute joint distribution
        prob_joint = utils.joint_distribution(data[columns_a + columns_b])

        # substract not part of thomas - need to ensure alignment
        # todo: should be part of thomas - submit pull-request to thomas
        prob_joint, prob_independent = prob_joint.extend_and_reorder(prob_joint, prob_independent)
        l1_distance = 0.5 * np.sum(np.abs(prob_joint.values - prob_independent.values))
        return l1_distance

    def save(self, path):
        """
        Save this synthesizer instance to the given path using pickle.

        Parameters
        ----------
        path: str
            Path where the synthesizer instance is saved.
        """
        # todo issue can't save if model is fitted - likely error within thomas
        if hasattr(self, 'model_'):
            pb = self.copy()
            del pb.model_

        with open(path, 'wb') as output:
            dill.dump(pb, output)

    @classmethod
    def load(cls, path):
        """Load a synthesizer instance from specified path.
        Parameters
        ----------
        path: str
            Path where the synthesizer instance is saved.

        Returns
        -------
        synthesizer : class
            Returns synthesizer instance.
        """
        with open(path, 'rb') as f:
            pb = dill.load(f)

        # recreate model_ attribute based on fitted network and cpt's
        if hasattr(pb, 'cpt_'):
            pb.model_ = BayesianNetwork.from_CPTs('PrivBayes', pb.cpt_.values())
        return pb



class PrivBayesFix(PrivBayes):
    """Extension to PrivBayes class to allow user to fix pre-sampled columns. Can be used to generate additional items
    for an already released synthetic dataset.
    """

    def __init__(self, epsilon=1.0, theta_usefulness=4, epsilon_split=0.3,
                 score_function='R', network_init=None, n_cpus=None, verbose=True):
        super().__init__(epsilon=epsilon, theta_usefulness=theta_usefulness, epsilon_split=epsilon_split,
                         score_function=score_function, network_init=network_init, n_cpus=n_cpus, verbose=verbose)

    def fit(self, data):
        super().fit(data)
        return self

    def sample_remaining_columns(self, fixed_data):
        self._check_is_fitted()
        fixed_data = self._check_fixed_data(fixed_data)

        data_synth = self._generate_data(fixed_data)
        data_synth = self._check_output_data(data_synth)

        if self.verbose:
            print("\nRemaining synthetic columns generated")
        return data_synth

    def _generate_data(self, fixed_data):
        n_records = fixed_data.shape[0]
        data_synth = np.empty([n_records, len(self.columns_)], dtype=object)

        for i in range(n_records):
            if self.verbose:
                print("\r", end='')
                print('Number of records generated: {} / {}'.format(i + 1, n_records), end='', flush=True)
            record_init = fixed_data.loc[i].to_dict()
            record = self._sample_record(record_init)
            data_synth[i] = list(record.values())

        # numpy.array to pandas.DataFrame with original column ordering
        data_synth = pd.DataFrame(data_synth, columns=[c.attribute for c in self.network_])
        return data_synth

    def _sample_record(self, record_init):
        # assume X has columns with values that correspond to the first nodes in the network
        # that we would like to fix and condition for.
        record = record_init

        # sample remaining nodes after fixing for input X
        for col_idx, pair in enumerate(self.network_[len(record_init):]):
            node = self.model_[pair.attribute]

            # specify pre-sampled conditioning values
            node_cpt = node.cpt
            node_states = node.states

            if node.conditioning:
                parent_values = [record[p] for p in node.conditioning]
                node_probs = node_cpt[tuple(parent_values)]
            else:
                node_probs = node_cpt.values
            # sampled_node_value = np.random.choice(node_states, p=node_probs)
            sampled_node_value = random.choices(node_states, weights=node_probs)
            record[node.name] = sampled_node_value
        return record

    def _check_fixed_data(self, data):
        """Checks whether the columns in fixed data where also seen in fit"""
        if isinstance(data, pd.Series):
            data = data.to_frame()

        if not np.all([c in self.columns_ for c in data.columns]):
            raise ValueError('Columns in fixed data not seen in fit.')
        if set(data.columns) == set(self.columns_):
            raise ValueError('Fixed data already contains all the columns that were seen in fit.')

        if self.verbose:
            sample_columns = [c for c in self.columns_ if c not in data.columns]
            print('Columns sampled and added to fixed data: {}'.format(sample_columns))
        # prevent integer column indexing issues
        data.columns = data.columns.astype(str)
        # make all columns categorical.
        data = data.astype('category')
        return data

class PrivBayesNP(PrivBayes):
    """Privbayes class with infinite-differential privacy, while still using epsilon value to limit the size of the network
    """

    def __init__(self, epsilon=1, theta_usefulness=4, epsilon_split=0.3,
                 score_function='R', network_init=None, n_cpus=None, verbose=True):
        self.epsilon1 = float(np.inf)
        self.epsilon2 = epsilon
        super().__init__(epsilon=self.epsilon1, theta_usefulness=theta_usefulness, epsilon_split=epsilon_split,
                         score_function=score_function, network_init=network_init, n_cpus=n_cpus, verbose=verbose)


    def _max_domain_size(self, data, node):
        """Computes the maximum domain size a node can have to satisfy theta-usefulness"""
        node_cardinality = utils.cardinality(data[node])
        max_domain_size = self.n_records_fit_ * self.epsilon2 / \
                          (2 * len(self.columns_) * self.theta_usefulness * node_cardinality)
        return max_domain_size


if __name__ == "__main__":
    data_path = '../../examples/data/original/adult.csv'
    data = pd.read_csv(data_path, engine='python')
    # columns = ['age', 'sex', 'education', 'workclass', 'income']
    # data = data.loc[:, columns]
    print(data.head())

    pb = PrivBayes(epsilon=1, n_cpus=4)
    pb.fit(data)
    df_synth = pb.sample(1000)
    pb.score(data, df_synth, score_dict=True)

    """test pb with init network"""
    # init_network = [AP_pair(node='age', parents=None),
    #                 AP_pair(node='education', parents=('age',)),
    #                 AP_pair(node='sex', parents=('age', 'education'))]
    #
    # pbinit = PrivBayes()
    # pbinit.set_network(init_network)
    # pbinit.fit(df)
    # df_synth_init = pbinit.sample(1000)

    # fixing a network - specify init network to fix those variables when generating
    pbfix = PrivBayesFix(epsilon=1, n_cpus=4)
    init_network = [APPair('age', None), APPair('education', ('age',))]
    # init_network = [APPair(node='age', parents=None),
    #                 APPair(node='education', parents=('age',)),
    #                 APPair(node='sex', parents=('age', 'education')),
    #                 APPair(node='workclass', parents=('age', 'education')),
    #                 APPair(node='income', parents=('sex', 'age'))]
    pbfix.set_network(init_network)
    pbfix.fit(data)
    df_synth_remaining = pbfix.sample_remaining_columns(df_synth[['age', 'education']])


    """test scoring functions"""
    pair = pb.network_[3]
    pb.mi_score(data, pair.attribute, pair.parents)
    pb.mi_score_thomas(data, pair.attribute, pair.parents)
    pb.r_score(data, pair.attribute, pair.parents)




"""Algorithms for generating synthetic data"""

from synthesis.synthesizers.marginal import MarginalSynthesizer, UniformSynthesizer
from synthesis.synthesizers.contingency import ContingencySynthesizer
from synthesis.synthesizers.privbayes import PrivBayes, PrivBayesFix

__all__ = [
    'MarginalSynthesizer',
    'UniformSynthesizer',
    'ContingencySynthesizer',
    'PrivBayes',
    'PrivBayesFix'
]


"""Methods for de-identifying original data"""

import numpy as np
import pandas as pd
import pandas_flavor as pf

@pf.register_dataframe_method
def replace_unique_values(df: pd.DataFrame, column_name: str, replace_string: str ='entity_') -> pd.DataFrame:
    """Replaces unique values with numbered string. Retains nans."""
    df[column_name] = df[column_name].astype(str)
    replace_dict = {v: (replace_string + str(i) if v != 'nan' else np.nan) for i, v  in
                    enumerate(df[column_name].unique())}
    df[column_name] = df[column_name].replace(replace_dict)
    return df

@pf.register_dataframe_method
def group_rare_values(df: pd.DataFrame, column_names=None, threshold=0.05, rare_group_label='other'):
    """For each column in dataframe replaces values that occur less than threshold"""
    column_names = column_names or df.columns
    if isinstance(column_names, str):
        column_names = [column_names]

    df = df.copy()
    for c in column_names:
        df.loc[:, c] = df[c].mask(df[c].map(df[c].value_counts(normalize=True)) < threshold, rare_group_label)
    return df

@pf.register_dataframe_method
def add_fake_identifiers(df, colummn_name='id',  id_string='FAKE_PERSON_'):
    """Add column with fake identifiers"""
    df = df.copy()
    df[colummn_name] = [id_string + str(i) for i in np.arange(1, df.shape[0]+1)]
    return df





"""Methods to determine whether the synthetic data performs similar to the original data on specific tasks"""

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from lifelines import KaplanMeierFitter
from sklearn.ensemble import RandomForestClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder
from sklearn.impute import SimpleImputer
from sklearn.model_selection import GridSearchCV
from sklearn.compose import ColumnTransformer
from sklearn.metrics import roc_auc_score, plot_roc_curve
from sklearn.base import BaseEstimator, TransformerMixin

from synthesis.evaluation._base import BaseMetric, BasePredictiveMetric, COLOR_PALETTE


class KaplanMeierComparison(BaseMetric):

    def __init__(self, time_column, event_column, group_column, labels=None):
        super().__init__(labels=labels)
        self.time_column = time_column
        self.event_column = event_column
        self.group_column = group_column

    def _fit(self, data):
        return {
            'time': data[self.time_column].astype(float),
            'event': data[self.event_column].astype(float),
            'group': data[self.group_column]
        }

    def fit(self, data_original, data_synthetic):
        self.stats_original_ = self._fit(data_original)
        self.stats_synthetic_ = self._fit(data_synthetic)
        return self

    def score(self):
        """for now only support plotting -
        could check whether logranktest on original and synthetic for each group agree on significance"""
        return None

    def plot(self):
        """
        Plot side-by-side kaplan-meier of input datasets
        """

        figsize = (10, 5)
        fig, ax = plt.subplots(1, 2, figsize=figsize, sharey=True)

        # sns.set(font_scale=1.5)
        # sns.despine()
        palette = ['#0d3d56', '#006887', '#0098b5', '#00cbde', '#00ffff']

        datasets = [self.stats_original_, self.stats_synthetic_]
        for data, label, ax_cur in zip(datasets, self.labels, ax):
            t = data['time']
            e = data['event']

            kmf = KaplanMeierFitter()
            groups = np.sort(data['group'].unique())
            for g, color in zip(groups, palette):
                mask = (data['group'] == g)
                kmf.fit(t[mask], event_observed=e[mask], label=g)
                ax_cur = kmf.plot_survival_function(ax=ax_cur, color=color)
                ax_cur.legend(title=self.group_column)
                ax_cur.set_title('Kaplan-Meier - {} data'.format(label))
                ax_cur.set_ylim(0, 1)
        plt.tight_layout()


class EnsureConsistentType(BaseEstimator, TransformerMixin):
    """Ensure consistent type in dataset - used to avoid issues with mixed types in columns"""
    def __init__(self, dtype=str):
        self.dtype=dtype

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return X.astype(self.dtype)


class TrainBothTestOriginalHoldout(BasePredictiveMetric):

    def __init__(self, labels=None, exclude_columns=None, y_column=None, random_state=None, n_jobs=None):
        super().__init__(labels=labels, exclude_columns=exclude_columns, astype_cat=False, y_column=y_column,
                         random_state=random_state, n_jobs=n_jobs)
        # todo make generic, enter any classifier/params but also provide default option

    def fit(self, data_original, data_synthetic):
        data_original, data_synthetic = self._check_input_data(data_original, data_synthetic)

        if self.y_column is None:
            self.y_column = data_original.columns[-1]

        X_original, y_original = self._split_xy(data_original)
        X_synthetic, y_synthetic = self._split_xy(data_synthetic)

        self.stats_original_ = self._fit(X_original, y_original)
        self.stats_synthetic_ = self._fit(X_synthetic, y_synthetic)

    def _fit(self, X, y):
        categorical_features = X.select_dtypes(include=['object']).columns
        numeric_features = [feat for feat in X.columns if not feat in categorical_features]

        preprocessor_num = Pipeline(steps=[
            ('simple_imputer_num', SimpleImputer(strategy='median')),
            ('num_scaling', MinMaxScaler()),
        ])

        preprocessor_cat = Pipeline(steps=[
            ('simple_imputer_cat', SimpleImputer(strategy='most_frequent')),
            ('ensure_consistent_dtype', EnsureConsistentType(dtype=str)),  # avoid mixed type columns
            ('categorical_encoding', OneHotEncoder(handle_unknown='ignore'))
        ])

        preprocessor = ColumnTransformer(transformers=[
            ('num', preprocessor_num, numeric_features),
            ('cat', preprocessor_cat, categorical_features)
        ])


        # Classifier
        clf_rf = RandomForestClassifier(class_weight='balanced', min_samples_leaf=0.05, random_state=self.random_state)

        # Pipeline
        pipe = Pipeline([('preprocessor', preprocessor),
                         ('classifier', clf_rf)])

        # Grid search
        params = {'classifier__n_estimators': [100, 150, 200],
                  'classifier__criterion': ['entropy', 'gini'],
                  'classifier__max_depth': [3, 5, 10],
                  'classifier__max_features': ['sqrt', 'log2']}

        grid_rf = GridSearchCV(pipe, param_grid=params, scoring='roc_auc', refit=True, cv=3, verbose=2,
                               n_jobs=self.n_jobs)
        grid_rf.fit(X, y)
        return grid_rf

    def score(self, data_original_test):
        X_test, y_test = self._split_xy(data_original_test)

        scores = {
            "roc_auc_original": roc_auc_score(y_test, self.stats_original_.predict_proba(X_test)[:, 1]),
            "roc_auc_synthetic": roc_auc_score(y_test, self.stats_synthetic_.predict_proba(X_test)[:, 1])
        }
        return scores

    def plot(self, data_original_test):
        """"Plot ROC-AOC Curves of both original and synthetic in single figure"""
        X_test, y_test = self._split_xy(data_original_test)

        fig, ax = plt.subplots()
        plot_roc_curve(self.stats_original_, X_test, y_test, ax=ax, name=self.labels[0], color=COLOR_PALETTE[0])
        plot_roc_curve(self.stats_synthetic_, X_test, y_test, ax=ax, name=self.labels[1], color=COLOR_PALETTE[1])
        ax.plot([0, 1], [0, 1], linestyle='--', lw=1, color='black', alpha=.8)
        plt.title('ROC Curve')
        plt.show()

    def _split_xy(self, data):
        y = data[self.y_column]
        X = data.drop(self.y_column, axis=1)
        return X, y


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['synthesis',
 'synthesis.evaluation',
 'synthesis.synthesizers',
 'synthesis.transformers']

package_data = \
{'': ['*']}

install_requires = \
['diffprivlib==0.5.1',
 'dill>=0.3.4,<0.4.0',
 'dython>=0.6.7,<0.7.0',
 'joblib>=1.0.1,<2.0.0',
 'lifelines>=0.26.0,<0.27.0',
 'matplotlib>=3.4.3,<4.0.0',
 'numpy>=1.21.2,<2.0.0',
 'pandas>=1.3.2,<2.0.0',
 'pip>=22.1.1,<23.0.0',
 'pyjanitor>=0.21.2,<0.22.0',
 'scikit-learn==1.0.2',
 'scipy>=1.7.1,<2.0.0',
 'seaborn>=0.11.2,<0.12.0',
 'thomas-core>=0.1.3,<0.2.0']

setup_kwargs = {
    'name': 'synthetic-data-generation',
    'version': '0.1.6',
    'description': 'Algorithms for generating synthetic data',
    'long_description': None,
    'author': 'Daan Knoors',
    'author_email': 'd.knoors@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



"""Module with metrics for comparison of datasets"""

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

from scipy.spatial.distance import jensenshannon
from dython.nominal import associations
from itertools import combinations

import synthesis.synthesizers.utils as utils
from synthesis.evaluation._base import BaseMetric, COLOR_PALETTE


class MarginalComparison(BaseMetric):

    def __init__(self, labels=None, exclude_columns=None, normalize=True):
        super().__init__(labels=labels, exclude_columns=exclude_columns)
        self.normalize = normalize

    def fit(self, data_original, data_synthetic):
        data_original, data_synthetic = self._check_input_data(data_original, data_synthetic)
        self.stats_original_ = {}
        self.stats_synthetic_ = {}
        self.stats_ = {}
        for c in data_original.columns:
            # compute value_counts for both original and synthetic - align indexes as certain column
            # values in the original data may not have been sampled in the synthetic data
            self.stats_original_[c], self.stats_synthetic_[c] = \
                data_original[c].value_counts(dropna=False, normalize=self.normalize).align(
                    data_synthetic[c].value_counts(dropna=False, normalize=self.normalize),
                    join='outer', axis=0, fill_value=0
                )
            self.stats_[c] = jensenshannon(self.stats_original_[c], self.stats_synthetic_[c])

        return self

    def score(self):
        average_js_distance = sum(self.stats_.values()) / len(self.stats_.keys())
        return average_js_distance

    def plot(self):
        column_names = self.stats_original_.keys()
        fig, ax = plt.subplots(len(column_names), 1, figsize=(8, len(column_names) * 4))

        for idx, col in enumerate(column_names):

            column_value_counts_original = self.stats_original_[col]
            column_value_counts_synthetic = self.stats_synthetic_[col]

            bar_position = np.arange(len(column_value_counts_original.values))
            bar_width = 0.35

            # with small column cardinality plot original distribution as bars, else plot as line
            if len(column_value_counts_original.values) <= 25:
                ax[idx].bar(x=bar_position, height=column_value_counts_original.values,
                            color=COLOR_PALETTE[0], label=self.labels[0], width=bar_width)
            else:
                ax[idx].plot(bar_position + bar_width, column_value_counts_original.values, marker='o',
                             markersize=3, color=COLOR_PALETTE[0], linewidth=2, label=self.labels[0])

            # synthetic distribution
            ax[idx].bar(x=bar_position + bar_width, height=column_value_counts_synthetic.values,
                        color=COLOR_PALETTE[1], label=self.labels[1], width=bar_width)

            ax[idx].set_xticks(bar_position + bar_width / 2)
            if len(column_value_counts_original.values) <= 25:
                ax[idx].set_xticklabels(column_value_counts_original.keys(), rotation=25)
            else:
                ax[idx].set_xticklabels('')
            # ax[idx].set_title('Column: ' +  r"$\bf{" + col + "}$" +
            #                   ' ~ jensen-shannon distance: ' + '{:.2f}'.format(self.stats_[col]))
            title = r"$\bf{" + col.replace('_', '\_') + "}$" + "\n jensen-shannon distance: {:.2f}".format(self.stats_[col])
            ax[idx].set_title(title)
            if self.normalize:
                ax[idx].set_ylabel('Probability')
            else:
                ax[idx].set_ylabel('Count')

            ax[idx].legend()
        fig.tight_layout()
        plt.show()

class AssociationsComparison(BaseMetric):

    def __init__(self, labels=None, exclude_columns=None, nom_nom_assoc='theil', nominal_columns='auto'):
        super().__init__(labels=labels, exclude_columns=exclude_columns, astype_cat=False)
        self.nom_nom_assoc = nom_nom_assoc
        self.nominal_columns = nominal_columns

    def fit(self, data_original, data_synthetic):
        data_original, data_synthetic = self._check_input_data(data_original, data_synthetic)

        self.stats_original_ = associations(data_original, nom_nom_assoc=self.nom_nom_assoc,
                                            nominal_columns=self.nominal_columns, nan_replace_value='nan',
                                            compute_only=True)['corr']
        self.stats_synthetic_ = associations(data_synthetic, nom_nom_assoc=self.nom_nom_assoc,
                                             nominal_columns=self.nominal_columns, nan_replace_value='nan',
                                             compute_only=True)['corr']
        return self

    def score(self):
        pairwise_correlation_distance = np.linalg.norm(self.stats_original_-self.stats_synthetic_, 'fro')
        return pairwise_correlation_distance

    def plot(self):
        pcd = self.score()

        fig, ax = plt.subplots(1, 2, figsize=(12, 10))
        cbar_ax = fig.add_axes([.91, 0.3, .01, .4])

        cmap = sns.diverging_palette(220, 10, as_cmap=True)

        # Original
        heatmap_original = sns.heatmap(self.stats_original_, ax=ax[0], square=True, annot=False, center=0, linewidths=0,
                         cmap=cmap, xticklabels=True, yticklabels=True, cbar_kws={'shrink': 0.8},
                         cbar_ax=cbar_ax, fmt='.2f')
        ax[0].set_title(self.labels[0] + '\n')

        # Synthetic
        heatmap_synthetic = sns.heatmap(self.stats_synthetic_, ax=ax[1], square=True, annot=False, center=0, linewidths=0,
                         cmap=cmap, xticklabels=True, yticklabels=False, cbar=False, cbar_kws={'shrink': 0.8})
        ax[1].set_title(self.labels[1] + '\n' + 'pairwise correlation distance: {}'.format(round(pcd, 4)))

        cbar = heatmap_original.collections[0].colorbar
        cbar.ax.tick_params(labelsize=10)
        # fig.tight_layout()
        plt.show()


class JointDistributionComparison(BaseMetric):

    def __init__(self, labels=None, exclude_columns=None, normalize=True, n_variables=2):
        super().__init__(labels=labels, exclude_columns=exclude_columns)
        self.normalize = normalize
        self.n_variables = n_variables

    def fit(self, data_original, data_synthetic):
        data_original, data_synthetic = self._check_input_data(data_original, data_synthetic)
        self.stats_original_ = {}
        self.stats_synthetic_ = {}
        self.stats_ = {}

        variable_combinations = combinations(data_original.columns, self.n_variables)
        for vars in variable_combinations:
            # compute joint distributions and convert Factors to series
            jt_original = utils.compute_distribution(data_original.loc[:, vars]).as_series()
            jt_synthetic = utils.compute_distribution(data_synthetic.loc[:, vars]).as_series()

            # align both original and synthetic in case values are missing in either dataset
            jt_original, jt_synthetic = jt_original.align(jt_synthetic, join='outer', fill_value=0, axis=0)
            self.stats_original_[vars] = jt_original.sort_index()
            self.stats_synthetic_[vars] = jt_synthetic.sort_index()

            # convert Factor to pd.Series and compute jensen-shannon distance
            self.stats_[vars] = jensenshannon(jt_original, jt_synthetic)

        return self

    def score(self):
        average_js_distance = sum(self.stats_.values()) / len(self.stats_.keys())
        return average_js_distance


"""Metrics for evaluating synthetic data"""

from synthesis.evaluation.metrics import MarginalComparison, AssociationsComparison
from synthesis.evaluation.efficacy import KaplanMeierComparison, TrainBothTestOriginalHoldout
from synthesis.evaluation.evaluator import SyntheticDataEvaluator, OriginalDataEvaluator

__all__ = [
    'MarginalComparison',
    'AssociationsComparison',
    'KaplanMeierComparison',
    'TrainBothTestOriginalHoldout',
    'SyntheticDataEvaluator',
    'OriginalDataEvaluator'
]



"""Base classes for all metrics"""
from abc import ABC, abstractmethod

from synthesis.synthesizers.utils import astype_categorical

COLOR_PALETTE = ['#393e46', '#ff5722', '#d72323']

class BaseMetric(ABC):

    def __init__(self, labels=None, exclude_columns=None, astype_cat=True):
        self.labels = labels or ['original', 'synthetic']
        self.exclude_columns = exclude_columns
        self.astype_cat = astype_cat

    def fit(self, data_original, data_synthetic):
        raise NotImplementedError("Implement fit method")

    def score(self, *args, **kwargs):
        raise NotImplementedError("Implement score method")

    def plot(self, *args, **kwargs):
        pass

    def _check_input_data(self, data_original, data_synthetic):
        # prevent integer column_name indexing issues
        data_original.columns = data_original.columns.astype(str)
        data_synthetic.columns = data_synthetic.columns.astype(str)

        # exclude columns
        if self.exclude_columns:
            columns = [c for c in data_original.columns if c not in self.exclude_columns]
            data_original = data_original[columns]
            data_synthetic = data_synthetic[columns]

        # ensure same columns order original data
        data_synthetic = data_synthetic[data_original.columns]

        # convert to categorical and add 'nan' as category
        if self.astype_cat:
            data_original = astype_categorical(data_original, include_nan=True)
            data_synthetic = astype_categorical(data_synthetic, include_nan=True)

        # todo check data alignment, i.e. whether synthetic data has same columns and categories as original
        # todo warn if dataset does not have same dimensions
        # todo convert column_name dtypes synthetic to original dtypes - warn user.
        return data_original, data_synthetic


class BasePredictiveMetric(BaseMetric):

    def __init__(self,  labels=None, exclude_columns=None, astype_cat=True, y_column=None, random_state=None, n_jobs=None):
        super().__init__(labels=labels, exclude_columns=exclude_columns, astype_cat=astype_cat)
        self.y_column = y_column
        self.random_state = random_state
        self.n_jobs = n_jobs

    def score(self, data_original_test):
        raise NotImplementedError("Implement score method")